function phaseout = osimOptimalTrackingContinuous(input)

% Define persistent variables so that splines are not conducted on each
% continuous function call
persistent mkrData grfr grfl ActsExp idData idTorques mcr mcl q

% rescale everything back to original
nPts = length(input.phase.time);
onesCol = ones(nPts,1);

% inputorig = input;

input.phase.time = (input.phase.time)*(input.auxdata.tMax-input.auxdata.tMin)+(input.auxdata.tMax+input.auxdata.tMin)/2;
input.phase.state = (input.phase.state).*(onesCol*(input.auxdata.stateMax-input.auxdata.stateMin))+onesCol*(input.auxdata.stateMax+input.auxdata.stateMin)/2;
input.phase.control = (input.phase.control).*(onesCol*(input.auxdata.controlMax-input.auxdata.controlMin))+onesCol*(input.auxdata.controlMax+input.auxdata.controlMin)/2;
if input.auxdata.calibrate
    input.phase.parameter = (input.phase.parameter(1,:)).*(input.auxdata.paramMax-input.auxdata.paramMin)+(input.auxdata.paramMax+input.auxdata.paramMin)/2;
end


%% Load parameters from the auxilary data
tScale = input.auxdata.tScale;

mu_s = 0; % Coefficient of static friction for foot-ground contact
mu_d = 1;%input.phase.parameter(1,9);%input.auxdata.mu;%parameters(1,1); % Coefficient of dynamic friction
mu_v = 0; % Coefficient of viscous friction
if input.auxdata.calibrate
    kValFits = input.phase.parameter(1,1:6);%input.auxdata.Kval; % Spring stiffnesses for foot ground contact
    cVal = input.auxdata.cVal; % Spring damping
    yVal = input.phase.parameter(1,7:8);
else
    kVal = input.auxdata.kVal;%input.auxdata.Kval; % Spring stiffnesses for foot ground contact
    cVal = input.auxdata.cVal; % Spring damping
    yVal = input.auxdata.yVal;
end

springMat = input.auxdata.springMat;
springBodyMat = input.auxdata.springBodyMat;
% ecr = input.auxdata.ecr;
% ecl = input.auxdata.ecl;
numSpringsBody = input.auxdata.numSpringsBody;

% SpringMat(7:end,2) = SpringMat(7:end,2)+Yval(1);
springMat(7:6+input.auxdata.numSprings/2,2) = springMat(7:6+input.auxdata.numSprings/2,2)+yVal(1);
springMat(7+input.auxdata.numSprings/2:end,2) = springMat(7+input.auxdata.numSprings/2:end,2)+yVal(2);

% Transition velocity between static and dynamic friction
latchvel = input.auxdata.latchvel;

muscParams.lmo = input.auxdata.lmo;
muscParams.lts = input.auxdata.lts;
muscParams.fMax = input.auxdata.fMax;
muscParams.pennAngle = input.auxdata.pennAngle;
muscParams.etmCoefs = input.auxdata.etmCoefs;
muscParams.nFrames = nPts;
muscParams.muscRef = input.auxdata.muscRef;

%% Calculate Spring Constant Values
if input.auxdata.calibrate
    kVal = zeros(1,input.auxdata.numSprings);

    springsPoly_l = input.auxdata.springsPoly_l;
    springsPoly_r = input.auxdata.springsPoly_r;

    springFitsMatrix_r = [ones(length(springsPoly_r),1) springsPoly_r(:,1) ...
        springsPoly_r(:,3) springsPoly_r(:,1).*springsPoly_r(:,3) ...
        springsPoly_r(:,1).^2 springsPoly_r(:,3).^2];

    springFitsMatrix_l = [ones(length(springsPoly_l),1) springsPoly_l(:,1) ...
        -springsPoly_l(:,3) -springsPoly_l(:,1).*springsPoly_l(:,3) ...
        springsPoly_l(:,1).^2 springsPoly_l(:,3).^2];

    kVal(1:numSpringsBody(1)+numSpringsBody(2)) = springFitsMatrix_r*kValFits(1:6)';
    kVal(numSpringsBody(1)+numSpringsBody(2)+1:end) = springFitsMatrix_l*kValFits(1:6)';
end

%% Initialize Variables

t = input.phase(1).time;

if ~exist('mkrData','var')
    q = fnval(input.auxdata.spq,t)';
    mkrData = fnval(input.auxdata.spMkrData,t)';
    grfr = fnval(input.auxdata.spGRFR,t)';
    grfl = fnval(input.auxdata.spGRFL,t)';
%     idData = fnval(input.auxdata.spID,t)';
    idTorques = fnval(input.auxdata.spTorques,t)';
    mcr = fnval(input.auxdata.spMcr,t)';
    mcl = fnval(input.auxdata.spMcl,t)';
elseif size(mkrData,1) ~= nPts
    q = fnval(input.auxdata.spq,t)';
    mkrData = fnval(input.auxdata.spMkrData,t)';
    grfr = fnval(input.auxdata.spGRFR,t)';
    grfl = fnval(input.auxdata.spGRFL,t)';
%     idData = fnval(input.auxdata.spID,t)';
    idTorques = fnval(input.auxdata.spTorques,t)';
    mcr = fnval(input.auxdata.spMcr,t)';
    mcl = fnval(input.auxdata.spMcl,t)';
end

%% Kinematics

% Set states
x = input.phase(1).state(:,1:31);
xp = input.phase(1).state(:,32:62);

% Calculate spring positions and velocities
[springPos, springVels] = opensimPointKin_v3(t,[x xp],springMat',springBodyMat);

% Store location of heel and toes body for tranferring GRFs
BodyPositions = springPos(:,1:12);
% mcr = SpringPos(:,13:15);
% mcl = SpringPos(:,16:18);
springPos(:,1:18) = [];
springVels(:,1:18) = [];

phaseout.springPos = springPos;
phaseout.springVels = springVels;

mcr(:,2) = 0;
mcl(:,2) = 0;

%% Calculate Ground Reaction Forces

% Calculate spring forces and moments about electrical centers
[footGRFsRMC, footGRFsLMC, elementForces] = calcGroundReactions_v3(springPos, springVels, ...
    numSpringsBody, mu_s, mu_d, mu_v, kVal, cVal, mcr, mcl, latchvel, tScale,input.auxdata.beltSpeed);

phaseout.elementForces = elementForces;

% Transfer moments to heel and toe origin for input into full body model
% This is necessary because the contact model ground reactions are
% calculated about some point in the room (force plate electrical centers),  
% and the model applies the loads at the origin of the foot and toes body
footGRFsR = footGRFsRMC;
footGRFsL = footGRFsLMC;

% Transfer to heel
footGRFsR(:,7:9) = cross(mcr-BodyPositions(:,1:3),footGRFsRMC(:,1:3),2)+footGRFsRMC(:,7:9);
footGRFsL(:,7:9) = cross(mcl-BodyPositions(:,7:9),footGRFsLMC(:,1:3),2)+footGRFsLMC(:,7:9);

% Transfer to toe
footGRFsR(:,10:12) = cross(mcr-BodyPositions(:,4:6),footGRFsRMC(:,4:6),2)+footGRFsRMC(:,10:12);
footGRFsL(:,10:12) = cross(mcl-BodyPositions(:,10:12),footGRFsLMC(:,4:6),2)+footGRFsLMC(:,10:12);

%% Calculate Reaction Loads on the Hands

if input.auxdata.handLoadControls
    handLoadsR = input.phase.control(:,(input.auxdata.nJerkControls+1):(input.auxdata.nJerkControls+6));
    handLoadsL = input.phase.control(:,(input.auxdata.nJerkControls+7):(input.auxdata.nJerkControls+12));
    
else
    % 3 linear springs, 3 torsion springs per hand
    % need linear displacement and angular displacement
    if input.auxdata.calibrate
        handSpring_k = input.phase.parameter(9:11);
        handSpring_tk = input.phase.parameter(12:14);
        if input.auxdata.springPosParams
            medianHandPos_r = input.auxdata.medianHandPos_r + input.phase.parameter(15:17);
            medianHandAng_r = input.auxdata.medianHandAng_r + input.phase.parameter(18:20);
            medianHandPos_l = input.auxdata.medianHandPos_l + input.phase.parameter(21:23);
            medianHandAng_l = input.auxdata.medianHandAng_l + input.phase.parameter(24:26);
        else
            medianHandPos_r = input.auxdata.medianHandPos_r;
            medianHandPos_l = input.auxdata.medianHandPos_l;
            medianHandAng_r = input.auxdata.medianHandAng_r;
            medianHandAng_l = input.auxdata.medianHandAng_l;
        end
    else
        handSpring_k = input.auxdata.handSpring_k;
        handSpring_tk = input.auxdata.handSpring_tk;
    end
    % handSpring_k = [1e3,1e4,1];
    % handSpring_tk = [0.1,0.1,100];

    handBodyMat = input.auxdata.handBodyMat;
    handForceLoc = input.auxdata.handForceLoc;
%     [handPos, handVel, handAng, handAngVel] = opensimGetBodyPosVel_OMP(t,[x xp],handBodyMat);
    [handPos, handVel, handAng, handAngVel] = opensimGetBodyPosVel_OMP(t,[x xp],handForceLoc,handBodyMat);
    handDisplacement_r = [-handPos(:,1) + medianHandPos_r(1),-handPos(:,2) + medianHandPos_r(2),-handPos(:,3) + medianHandPos_r(3)];
    handDisplacement_l = [-handPos(:,4) + medianHandPos_l(1),-handPos(:,5) + medianHandPos_l(2),-handPos(:,6) + medianHandPos_l(3)];
    handAngDisp_r = [-handAng(:,1) + medianHandAng_r(1),-handAng(:,2) + medianHandAng_r(2),-handAng(:,3) + medianHandAng_r(3)];
    handAngDisp_l = [-handAng(:,4) + medianHandAng_l(1),-handAng(:,5) + medianHandAng_l(2),-handAng(:,6) + medianHandAng_l(3)];

    handLoadsR = zeros(nPts,6); % There is no z component of the torque because the hands can rotate about the bar
    handLoadsL = zeros(nPts,6);

    handLoadsR(:,1) = (handDisplacement_r(:,1)*handSpring_k(1)).*(1+cVal(1)*handVel(:,1)); %TODO: Check this
    handLoadsR(:,2) = (handDisplacement_r(:,2)*handSpring_k(2)).*(1+cVal(1)*handVel(:,2));
    handLoadsR(:,3) = (handDisplacement_r(:,3)*handSpring_k(3)).*(1+cVal(1)*handVel(:,3));
    handLoadsR(:,4) = (handAngDisp_r(:,1)*handSpring_tk(1)).*(1+cVal(1)*handAngVel(:,1));
    handLoadsR(:,5) = (handAngDisp_r(:,2)*handSpring_tk(2)).*(1+cVal(1)*handAngVel(:,2));
    handLoadsR(:,6) = (handAngDisp_r(:,3)*handSpring_tk(3)).*(1+cVal(1)*handAngVel(:,3));

    handLoadsL(:,1) = handDisplacement_l(:,1)*handSpring_k(1).*(1+cVal(1)*handVel(:,4)); %TODO: Check this
    handLoadsL(:,2) = handDisplacement_l(:,2)*handSpring_k(2).*(1+cVal(1)*handVel(:,5));
    handLoadsL(:,3) = handDisplacement_l(:,3)*handSpring_k(3).*(1+cVal(1)*handVel(:,6));
    handLoadsL(:,4) = (handAngDisp_l(:,1)*handSpring_tk(1)).*(1+cVal(1)*handAngVel(:,4));
    handLoadsL(:,5) = (handAngDisp_l(:,2)*handSpring_tk(2)).*(1+cVal(1)*handAngVel(:,5));
    handLoadsL(:,6) = (handAngDisp_l(:,3)*handSpring_tk(3)).*(1+cVal(1)*handAngVel(:,6));
end

phaseout.handLoadsR = handLoadsR;
phaseout.handLoadsL = handLoadsL;


%% Dynamics

% Extract for input into full body dynamic model
torques = zeros(nPts,25);%Set to zero because they will be determined by inverse dynamics

% Formulate inputs into opensim (states and controls w/ GRFs)
states = [x xp];
appliedLoads = [torques footGRFsR footGRFsL handLoadsR handLoadsL];
% xpp = input.phase(1).control(:,1:29);

states(isnan(states)) = 0;
states(isinf(states)) = 0;

% Set accelerations
xpp = input.phase.state(:,63:end);
% Set jerk
xppp = input.phase.control(:,1:input.auxdata.nJerkControls);
if input.auxdata.extraControls
    controlTorques = input.phase.control(:,(input.auxdata.nJerkControls+1):end);
end


% Calculate state derivatives and marker locations
% [residualLoads, mkrPos] = Control_OpenSimMPI_muscle_ID([x xp], xpp, appliedLoads,t,[1 0 0]);
[residualLoads, mkrPos] = opensimID_OMP(t,[x xp], xpp, appliedLoads,[1 1 0]);
torques = residualLoads(:,7:end);

% Create new reordered marker matrix
mkrPosReorder = zeros(size(mkrPos,1),size(mkrPos,2)*size(mkrPos,3));
for iii = 1:size(mkrPos,2)
    mkrPosReorder(:,(iii-1)*3+1) = squeeze(mkrPos(:,iii,1));
    mkrPosReorder(:,(iii-1)*3+2) = squeeze(mkrPos(:,iii,2));
    mkrPosReorder(:,(iii-1)*3+3) = squeeze(mkrPos(:,iii,3));
end

% Set rescaled derivative of state vector
phaseout(1).dynamics = (input.auxdata.tMax-input.auxdata.tMin)*([xp xpp xppp])./(onesCol*(input.auxdata.stateMax-input.auxdata.stateMin));

if input.auxdata.extraControls
    phaseout(1).path = ([residualLoads(:,1:6),torques - controlTorques]-...
        ones(nPts,1)*(input.auxdata.pathMax+input.auxdata.pathMin)/2)...
        ./(ones(nPts,1)*(input.auxdata.pathMax-input.auxdata.pathMin));
else
    phaseout(1).path = ([residualLoads(:,1:6)]-...
        ones(nPts,1)*(input.auxdata.pathMax+input.auxdata.pathMin)/2)...
        ./(ones(nPts,1)*(input.auxdata.pathMax-input.auxdata.pathMin));
end


%% Cost Function

% Output GRFs split between segments used for creating plots after
% simulation finishes
phaseout.GRFsSplit = [footGRFsRMC footGRFsLMC];

% Calculate net GRFs at the electrical center for use in objective function
footGroundReactionsRMC = footGRFsRMC(:,[1:3 7:9])+footGRFsRMC(:,[4:6 10:12]);
footGroundReactionsLMC = footGRFsLMC(:,[1:3 7:9])+footGRFsLMC(:,[4:6 10:12]);

mkrError = (mkrPosReorder-mkrData);
grfError = [grfr-footGroundReactionsRMC grfl-footGroundReactionsLMC];
% activationError = activations-ActsExp;
if input.auxdata.extraControls
    torqueErrors = controlTorques-idTorques;
else
    torqueErrors = torques-idTorques;
end


% qError = x-q;
toeAngleError = x(:,[13,20]) - q(:,[13,20]);

% Set integrand, we penalize differences between model and experimental
% markers and differences between model and experimental ground reactions
% calculated at the electrical centers of the force plates, additionally we
% minimize the rate of change of the applied joint loads
if input.auxdata.handLoadControls
    phaseout(1).integrand = [mkrError...
        grfError(:,[1:3 7:9]) ...
        grfError(:,[4:6 10:12]) ...
        torqueErrors(:,[1:6,8:13])... % [1:6,8:13,15:end] We don't want to track arms and back
        toeAngleError ...
        [handLoadsR,handLoadsL]...
        xppp].^2;
else
    phaseout(1).integrand = [mkrError...
        grfError(:,[1:3 7:9]) ...
        grfError(:,[4:6 10:12]) ...
        torqueErrors(:,[1:6,8:13])... %  Old: [1:6,8:13,15:end] We don't want to track arms and back
        toeAngleError ...
        xppp].^2;
end
    
% Scale integrand
phaseout.integrand = (phaseout.integrand)./(onesCol*(input.auxdata.integralMax-input.auxdata.integralMin));

% phaseout.EtMmoments = EtMmoments;
% Set more outputs for post-simulation analyses
phaseout.mkrPos = mkrPos;
phaseout.markers = mkrPosReorder;
phaseout.torques = torques;
phaseout.kVal = kVal;
phaseout.yVal = yVal;
if ~input.auxdata.handLoadControls
    phaseout.handSpring_k = handSpring_k;
    phaseout.handSpring_tk = handSpring_tk;
end
end
